@@ -27,7 +27,7 @@ from .matches import *
 from .targets import *
 
 from .rule import ChainRule
-from .chain import Chain
+from .chain import Chain, BuiltinChain
 from .table import IptablesPacketFilterTable
 
 from .exceptions import (
