@@ -22,10 +22,9 @@ import traceback
 
 from typing import Callable, List, Optional
 
-from .exceptions import (
-        IptablesError, IptablesParsingError, IptablesExecutionError)
+from .exceptions import (IptablesError, IptablesParsingError)
 from .rule import ChainRule
-from .targets import ChainTarget
+from .targets import Target, Targets, ChainTarget
 from .deps import get_logger
 
 _logger = get_logger('linuxnet.iptables.chain')
@@ -38,26 +37,12 @@ class Chain:
     number (rule numbers start with 1).
     """
 
-    def __init__(self,          # pylint: disable=too-many-arguments
-                        chain_name: str,
-                        policy: Optional[str] =None,
-                        reference_count=0,
-                        packet_count=0, byte_count=0):
+    def __init__(self, chain_name: str):
         """
         :param chain_name: real chain name
-        :param policy: the chain policy if this is a builtin chain, ``None``
-            otherwise
-        :param reference_count: number of rules referencing this chain
-        :param packet_count: number of packets that flowed through this chain
-        :param byte_count: number of bytes that flowed through this chain
         """
         self.__real_chain_name = chain_name
-        self.__policy = policy
-        # The policy packet/byte count are the stats associated with
-        # the chain policy (applicable only for builtin chains)
-        self.__policy_packet_count = packet_count
-        self.__policy_byte_count = byte_count
-        self.__reference_count = reference_count
+        self.__reference_count = 0
         self.__packet_count = 0
         self.__byte_count = 0
         self.__rule_list = []
@@ -66,10 +51,10 @@ class Chain:
     def __str__(self):
         return f'Chain({self.__real_chain_name})'
 
-    def is_builtin(self) -> bool:
+    def is_builtin(self) -> bool:       # pylint: disable=no-self-use
         """Returns ``True`` if this is a built-in chain (e.g. ``INPUT``)
         """
-        return self.__policy is not None
+        return False
 
     def get_reference_count(self) -> int:
         """Returns the reference count of a (non-builtin) chain; returns 0
@@ -77,18 +62,6 @@ class Chain:
         """
         return self.__reference_count
 
-    def get_policy_packet_count(self) -> int:
-        """Returns the number of packets that were handled as per
-        the chain policy; returns 0 for non-builtin chains
-        """
-        return self.__policy_packet_count
-
-    def get_policy_byte_count(self) -> int:
-        """Returns the number of bytes that were handled as per
-        the chain policy; returns 0 for non-builtin chains
-        """
-        return self.__policy_byte_count
-
     def get_packet_count(self) -> int:
         """Returns the packet count of the chain
         """
@@ -111,12 +84,6 @@ class Chain:
             return self.__real_chain_name
         return self.__pft.rcn2lcn(self.__real_chain_name)
 
-    def get_policy(self) -> Optional[str]:
-        """Returns the policy of the (builtin) chain, or ``None`` if
-        this is not a builtin chain.
-        """
-        return self.__policy
-
     def has_unparsed_rules(self) -> bool:
         """Returns ``True`` if the chain contains unparsed rules
         """
@@ -135,11 +102,7 @@ class Chain:
         return count
 
     def get_rules(self) -> List[ChainRule]:
-        """Returns the chain rules.
-
-        The return value is a copy to avoid inadvertent modifications of
-        the internal rule list (since the internal rule list should reflect
-        the system's state).
+        """Returns a list that contains the chain rules.
         """
         return self.__rule_list[:]
 
@@ -174,12 +137,6 @@ class Chain:
                                 packet_count=rule.get_packet_count(),
                                 byte_count=rule.get_byte_count())
 
-    def _set_stats(self, packet_count: int, byte_count: int) -> None:
-        """Set the packet/byte counts of this chain
-        """
-        self.__packet_count = packet_count
-        self.__byte_count = byte_count
-
     def _update_stats(self, packet_count: int, byte_count: int) -> None:
         """Update the packet/byte counts of this chain
         """
@@ -188,9 +145,9 @@ class Chain:
 
     def find_rule_by_target_lcn(self,
                         logical_chain_name: str) -> List[ChainRule]:
-        """Return a list of rules with the specified target
+        """Return a list of rules with the specified chain as a target
 
-        :param logical_chain_name: identifies the chain targetted by the rule
+        :param logical_chain_name: identifies the chain targeted by the rule
         """
         rule_list = []
         for rule in self.__rule_list:
@@ -206,30 +163,40 @@ class Chain:
                 rule_list.append(rule)
         return rule_list
 
-    def find_rule_by(self, *, match=None, target=None) -> List[ChainRule]:
+    def find_rule_by(self, *, match: Optional['Match'] =None,
+                        is_only_match=True,
+                        target: Optional[Target] =None) -> List[ChainRule]:
         """Return a list of :class:`ChainRule` objects where the rule
         contains the specified ``match`` object or has the specified ``target``
-        (target comparison is by name), or both. If no ``match`` or ``target``
-        is present, an empty list is returned.
+        (target comparison is by name).
+        If both ``match`` and ``target`` are specified, returned rules
+        must satisfy both criteria.
+        If no ``match`` or ``target`` is present, an empty list is returned.
 
-        :param match: optional :class:`Match` object; use a :class:`MatchNone`
-            object to find a rule that has no matches
-        :param target: optional :class:`Target` object; use a
-            :class:`TargetNone` object to find a rule that has no target
+        :param match: :class:`Match` object to compare against;
+            if ``match`` is ``None``, do not perform any match comparisons;
+            if ``match`` is a :class:`MatchNone` object, this will
+            match a rule that has no matches
+        :param is_only_match: if ``True`` the specified ``match`` must
+            be the only match used in the rule
+        :param target: :class:`Target` object to compare against;
+            if ``target`` is ``None``, do not perform any target comparisons;
+            if ``target`` is a :class:`TargetNone` object, this will
+            match a rule that has no target
         """
         if match is None and target is None:
             return []
         return [rule for rule in self.__rule_list
-                    if (match is None or rule.has_match(match)) and
+            if (match is None or rule.has_match(match, is_only_match)) and
                         (target is None or rule.has_target(target))]
 
-    def get_pft(self):
+    def get_pft(self) -> 'IptablesPacketFilterTable':
         """Returns the :class:`IptablesPacketFilterTable` where this
         chain belongs
         """
         return self.__pft
 
-    def set_pft(self, pft) -> None:
+    def _set_pft(self, pft) -> None:
         """Set the :class:`IptablesPacketFilterTable` where this
         :class:`Chain` belongs.
 
@@ -237,7 +204,7 @@ class Chain:
         """
         self.__pft = pft
 
-    def clear_pft(self) -> None:
+    def _clear_pft(self) -> None:
         """Reset the :class:`IptablesPacketFilterTable` where this
         :class:`Chain` belongs.
         """
@@ -246,7 +213,13 @@ class Chain:
     def flush(self) -> None:
         """Delete all rules from this chain
         """
-        _ = self.__pft.iptables_run(['-F', self.__real_chain_name], check=True)
+        try:
+            args = ['-F', self.__real_chain_name]
+            _ = self.__pft.iptables_run(args, check=True)
+        except Exception:
+            _logger.exception("Failed to flush chain %s",
+                                self.__real_chain_name)
+            raise
         for rule in self.__rule_list:
             try:
                 self.__dec_target_refcount(rule.get_target())
@@ -256,6 +229,11 @@ class Chain:
                                         self.flush.__qualname__)
         self.__rule_list.clear()
 
+    def _setref(self, count: int) -> None:
+        """Set the chain reference count
+        """
+        self.__reference_count = count
+
     def _incref(self) -> None:
         """Increase the chain reference count
         """
@@ -270,7 +248,7 @@ class Chain:
             _logger.warning("Negative refcount for chain %s",
                                 self.__real_chain_name)
 
-    def __inc_target_refcount(self, target) -> None:
+    def __inc_target_refcount(self, target: Target) -> None:
         """If target is a :class:`ChainTarget`, increase the refcount of the
         corresponding chain.
         """
@@ -283,7 +261,7 @@ class Chain:
             _logger.warning("Missed refcount increase for chain %s",
                                 target.get_target_name())
 
-    def __dec_target_refcount(self, target) -> None:
+    def __dec_target_refcount(self, target: Target) -> None:
         """If target is a :class:`ChainTarget`, decrease the refcount of the
         corresponding chain.
 
@@ -309,18 +287,36 @@ class Chain:
             rule = self.__rule_list[i]
             rule._inc_rulenum()  # pylint: disable=protected-access
 
+    def __verify_no_owner(self, rule: ChainRule) -> None:
+        """Raise an exception is this rule has an owner
+        """
+        owner_chain = rule.get_chain()
+        if owner_chain is None:
+            return
+        if owner_chain is self:
+            raise IptablesError('rule already in this chain')
+        _logger.error("attempt to insert rule of chain '%s' to '%s'",
+                        owner_chain, self)
+        raise IptablesError('rule belongs to another chain')
+
     def append_rule(self, rule: ChainRule) -> None:
         """Append the new rule at the end of the chain
+
+        Raises an :exc:`IptablesError` if the rule is already part of a chain
         """
-        if rule.get_chain() is not None:
-            raise IptablesError('rule belongs to another chain')
+        self.__verify_no_owner(rule)
         rule_args = rule.to_iptables_args()
         if not rule_args:
             _logger.warning("%s: rule has no args: %s",
                 self.append_rule.__qualname__, rule)
             return
         args = ['-A', self.__real_chain_name] + rule_args
-        _ = self.__pft.iptables_run(args, check=True)
+        try:
+            _ = self.__pft.iptables_run(args, check=True)
+        except Exception:
+            _logger.exception("Failed to append rule to chain %s",
+                        self.__real_chain_name)
+            raise
         self.__rule_list.append(rule)
         self.__added_rule(rule, rulenum=len(self.__rule_list))
 
@@ -328,10 +324,13 @@ class Chain:
         """Insert the new rule at the beginning of the chain (by
         default) or as rule number ``rulenum``.
 
-        :param rulenum: rule number (starting with 1)
+        Raises an :exc:`IptablesError` if the rule is already part of a chain
+
+        :param rule: the :class:`ChainRule` to insert
+        :param rulenum: rule number (starting with 1) for the inserted
+            rule
         """
-        if rule.get_chain() is not None:
-            raise IptablesError('rule belongs to another chain')
+        self.__verify_no_owner(rule)
         rule_args = rule.to_iptables_args()
         if not rule_args:
             _logger.warning("%s: rule has no args: %s",
@@ -350,13 +349,21 @@ class Chain:
             _ = self.__pft.iptables_run(args, check=True)
         except Exception:
             self.__rule_list.pop(rule_index)
+            _logger.exception("Failed to insert rule to chain %s",
+                        self.__real_chain_name)
             raise
         self.__added_rule(rule, rulenum=rule_index+1)
 
     def delete_rule(self, rule: ChainRule) -> None:
-        """Delete the specified ``rule``: the rule must belong to this chain.
+        """Delete the specified ``rule``.
+
+        Raises an :exc:`IptablesError` if the rule is not part of this chain.
         """
-        if rule.get_chain() is not self:
+        owner_chain = rule.get_chain()
+        if owner_chain is not self:
+            _logger.error(
+                "attempt to delete rule of chain '%s' from chain '%s'",
+                    owner_chain, self)
             raise IptablesError('attempt to delete rule from wrong chain')
         rule_index = rule.get_rulenum() - 1
         if self.__rule_list[rule_index] is not rule:
@@ -368,7 +375,7 @@ class Chain:
     def delete_rulenum(self, rulenum: int) -> None:
         """Delete the rule with the specified rule number
 
-        Raises an :class:`IptablesError` if the number is invalid
+        Raises an :exc:`IptablesError` if the number is invalid
 
         :param rulenum: rule number (numbering starts from 1)
         """
@@ -388,12 +395,11 @@ class Chain:
         cmd = ['-D', self.__real_chain_name, f'{rulenum}']
         try:
             _ = self.__pft.iptables_run(cmd, check=True)
-        except Exception as ex:
+        except Exception:
             self.__rule_list.insert(rule_index, rule)
-            _logger.exception("Rule deletion failed")
-            raise IptablesExecutionError(
-                    f'unable to delete rule {rulenum} '
-                    f'from chain {self.get_real_name()}') from ex
+            _logger.exception("unable to delete rule %d from chain %s",
+                        rulenum, self.__real_chain_name)
+            raise
         rule._deleted()         # pylint: disable=protected-access
         self.__dec_target_refcount(rule.get_target())
         #
@@ -420,23 +426,28 @@ class Chain:
     def delete_rule_by_pred(self, pred: Callable[[ChainRule], bool]) -> int:
         """Delete all rules for which ``pred`` returns ``True``.
 
-        Returns the number of deleted rules
-
         :param pred: a ``Callable`` object
+        :rtype: number of deleted rules
         """
         deletion_list = [rule for rule in self.__rule_list if pred(rule)]
         return self.__delete_rules(deletion_list)
 
-    def delete_rule_if(self, *, match=None, target=None) -> int:
+    def delete_rule_if(self, *, match: Optional['Match'] =None,
+                                target: Optional[Target] =None) -> int:
         """Delete all rules with the specified ``match`` and/or ``target``.
         If no ``match`` or ``target`` is present, this is a no-op.
 
-        Returns the number of deleted rules.
-
-        :param match: optional :class:`Match` object; use a :class:`MatchNone`
-            object to delete a rule that has no matches
-        :param target: optional :class:`Target` object; use a
-            :class:`TargetNone` object to delete a rule that has no target
+        :param match: :class:`Match` object to compare against;
+            the comparison will be successful if this is the **only** match
+            used in the rule;
+            if ``match`` is ``None``, do not perform any match comparisons;
+            if ``match`` is a :class:`MatchNone` object, this will
+            match a rule that has no matches
+        :param target: :class:`Target` object to compare against;
+            if ``target`` is ``None``, do not perform any target comparisons;
+            if ``target`` is a :class:`TargetNone` object, this will
+            match a rule that has no target
+        :rtype: number of deleted rules
         """
         if match is None and target is None:
             return 0
@@ -446,9 +457,8 @@ class Chain:
     def delete_rule_by_target_chain(self, chain: 'Chain') -> int:
         """Delete all rules that jump/goto the specified chain.
 
-        Returns the number of deleted rules
-
         :param chain: a :class:`Chain` object
+        :rtype: number of deleted rules
         """
         return self.delete_rule_by_pred(pred=lambda r: r.targets_chain(chain))
 
@@ -460,8 +470,9 @@ class Chain:
         self.__pft.zero_counters(chain=self)
 
     @classmethod
-    def __parse_chain_line(cls, line) -> 'Chain':
+    def __parse_chain_line(cls, line: str) -> 'Chain':
         """Parse a line which has one of the following 2 forms:
+
             Chain INPUT (policy ACCEPT 9108340 packets, 10054611039 bytes)
             Chain host_origin (1 references)
 
@@ -485,7 +496,11 @@ class Chain:
         try:
             param_fields = fields[2][1:-1].split()
             if param_fields[0] == 'policy':
-                policy = param_fields[1]
+                target_name = param_fields[1]
+                policy = Targets.get_special(target_name)
+                if policy is None:
+                    raise IptablesParsingError(
+                        f"unknown policy target for chain {real_chain_name}")
                 if param_fields[3].startswith('packets'):
                     packet_count = int(param_fields[2])
                 if param_fields[5].startswith('bytes'):
@@ -495,8 +510,12 @@ class Chain:
             else:
                 _logger.warning("unable to parse line: %s", line)
                 return None
-            return Chain(real_chain_name, policy, reference_count,
+            if policy is not None:
+                return BuiltinChain(real_chain_name, policy,
                                         packet_count, byte_count)
+            chain = Chain(real_chain_name)
+            chain._setref(reference_count)
+            return chain
         except IndexError as idxerr:
             raise IptablesParsingError(
                         'insufficient number of fields', line=line) from idxerr
@@ -545,3 +564,72 @@ class Chain:
             rule_list.append(rule)
         chain._set_rule_list(rule_list) # pylint: disable=protected-access
         return chain
+
+
+class BuiltinChain(Chain):
+    """This class is used to represent an iptables built-in chain.
+
+    Instances of this class are not intended to be created by the user;
+    they are created when processing the output of **iptables(8)**
+    """
+
+    def __init__(self, chain_name, policy: Target,
+                        packet_count: int, byte_count: int):
+        """
+        :param chain_name: builtin chain name
+        :param policy: chain policy target
+        :param packet_count: number of packets that were processed according
+            to the policy
+        :param byte_count: number of bytes for packets that were processed
+            according to the policy
+        """
+        super().__init__(chain_name)
+        self.__policy = policy
+        self.__policy_packet_count = packet_count
+        self.__policy_byte_count = byte_count
+
+    def __str__(self):
+        return f'BuiltinChain({self.__real_chain_name})'
+
+    def is_builtin(self) -> bool:
+        """Always ``True``
+        """
+        return True
+
+    def get_policy(self) -> Target:
+        """Returns the policy target of this builtin chain
+        """
+        return self.__policy
+
+    def get_policy_packet_count(self) -> int:
+        """Returns the number of packets that were handled as per
+        the chain policy
+        """
+        return self.__policy_packet_count
+
+    def get_policy_byte_count(self) -> int:
+        """Returns the number of bytes that were handled as per
+        the chain policy
+        """
+        return self.__policy_byte_count
+
+    def _set_stats(self, packet_count: int, byte_count: int) -> None:
+        """Set the packet/byte counts of this chain
+        """
+        total_packet_count = packet_count + self.__policy_packet_count
+        total_byte_count = byte_count + self.__policy_byte_count
+        self._update_stats(total_packet_count, total_byte_count)
+
+    def set_policy(self, policy: Target) -> None:
+        """Set the policy target of this builtin chain
+        """
+        builtin = self.get_real_name()
+        policy_name = policy.get_target_name()
+        cmd = ['-P', builtin, policy_name]
+        try:
+            _ = self.get_pft().iptables_run(cmd, check=True)
+        except Exception:
+            _logger.exception("Failed to set policy of builtin chain %s",
+                        builtin)
+            raise
+        self.__policy = policy
