@@ -21,7 +21,7 @@
 from typing import List, Optional
 
 from .exceptions import IptablesError, IptablesParsingError
-from .matches import Match, MatchNone, PacketMatch
+from .matches import Match, MatchNone, PacketMatch, CommentMatch
 from .matches.match import MatchParser
 from .parsing import LookaheadIterator, RuleFieldIterator
 from .targets import Target, ChainTarget, UnparsedTarget, TargetNone
@@ -125,7 +125,7 @@ class ChainRule:        # pylint: disable=too-many-instance-attributes
         """
         return self.__byte_count
 
-    def get_chain(self):
+    def get_chain(self) -> 'Chain':
         """Returns the :class:`Chain` where this rule belongs (returns
         ``None`` if this rule is not in any chain)
         """
@@ -192,20 +192,30 @@ class ChainRule:        # pylint: disable=too-many-instance-attributes
         self.__target = target
 
     def get_match_list(self) -> List[Match]:
-        """Returns match list
+        """Returns the match list of this rule.
         """
-        return self.__match_list
+        return self.__match_list[:]
 
-    def has_match(self, match: Match) -> bool:
+    def has_match(self, match: Match, is_only_match=True) -> bool:
         """Returns ``True`` if the match list of this rule consists only
-        of the specified match. An object of :class:`MatchNone` can be
+        of the specified match
+        (when ``is_only_match`` is ``True``)
+        or if the match list contains the specified match
+        (when ``is_only_match`` is ``False``).
+
+        An object of :class:`MatchNone` can be
         used to test for an empty match list.
         """
         if isinstance(match, MatchNone):
             return not bool(self.__match_list)
-        if len(self.__match_list) != 1:
-            return False
-        return match == self.__match_list[0]
+        if is_only_match:
+            if len(self.__match_list) != 1:
+                return False
+            return match == self.__match_list[0]
+        for existing_match in self.__match_list:
+            if match == existing_match:
+                return True
+        return False
 
     def has_target(self, target: Target) -> bool:
         """Returns ``True`` if the rule has the specified target.
@@ -216,7 +226,7 @@ class ChainRule:        # pylint: disable=too-many-instance-attributes
             return self.__target is None
         return self.__target == target
 
-    def targets_chain(self, chain) -> bool:
+    def targets_chain(self, chain: 'Chain') -> bool:
         """Returns ``True`` if the target of this rule is the specified chain
 
         :param chain: a :class:`Chain` object
@@ -227,8 +237,8 @@ class ChainRule:        # pylint: disable=too-many-instance-attributes
         return (target_chain is not None and
                 target_chain.get_real_name() == chain.get_real_name())
 
-    def get_target_chain(self):
-        """Returns the (non-builtin) :class:`Chain` object that is
+    def get_target_chain(self) -> Optional['Chain']:
+        """Returns the :class:`Chain` object that is
         the target of this rule, or ``None`` if this rule does not
         target a chain.
         """
@@ -244,6 +254,14 @@ class ChainRule:        # pylint: disable=too-many-instance-attributes
             raise IptablesError("rule is not in IptablesPacketFilterTable")
         return self.__target.resolve_chain(pft)
 
+    def matches_all_packets(self) -> bool:
+        """Returns ``True`` iff this rule matches all packets
+        """
+        for match in self.__match_list:
+            if not isinstance(match, CommentMatch):
+                return False
+        return True
+
     def to_iptables_args(self) -> List[str]:
         """Returns a list suitable to be used as an argument to
         the **iptables(8)** command
@@ -263,7 +281,7 @@ class ChainRule:        # pylint: disable=too-many-instance-attributes
         return retval
 
     def jump_to(self, *, target: Optional[Target] =None,
-                                chain=None) -> 'ChainRule':
+                            chain: Optional['Chain'] =None) -> 'ChainRule':
         """Add a jump to the specified target.
         The target is identified either via the ``target`` argument or
         via the ``chain`` argument.
@@ -272,10 +290,9 @@ class ChainRule:        # pylint: disable=too-many-instance-attributes
             - both ``target`` and ``chain`` arguments are not ``None``
             - the rule is already part of a :class:`Chain`
 
-        Returns this object.
-
         :param target: optional :class:`Target` object
         :param chain: optional :class:`Chain` object
+        :rtype: this :class:`ChainRule` object
         """
         if self.__owner_chain is not None:
             rcn = self.__owner_chain.get_real_name()
@@ -287,15 +304,14 @@ class ChainRule:        # pylint: disable=too-many-instance-attributes
         self.__target = target
         return self
 
-    def go_to(self, *, chain) -> 'ChainRule':
+    def go_to(self, *, chain: 'Chain') -> 'ChainRule':
         """Add a goto to the specified chain.
 
         Raises an :exc:`IptablesError` if the rule is already part
         of a :class:`Chain`
 
-        Returns this object.
-
         :param chain: a :class:`Chain` object
+        :rtype: this :class:`ChainRule` object
         """
         if self.__owner_chain is not None:
             rcn = self.__owner_chain.get_real_name()
@@ -342,7 +358,8 @@ class ChainRule:        # pylint: disable=too-many-instance-attributes
 
     @classmethod
     def create_from_existing(cls,   # pylint: disable=too-many-locals
-                        iptables_output_line: str, pft) -> 'ChainRule':
+                        iptables_output_line: str,
+                        pft: 'IptablesPacketFilterTable') -> 'ChainRule':
         """Create a ChainRule from a line of ``iptables -xnv`` output
 
         :param iptables_output_line: line of ``iptables -xnv`` output
