@@ -26,8 +26,7 @@ from typing import Dict, Iterator, List, Optional, Tuple
 
 from .exceptions import (
                 IptablesError, IptablesParsingError, IptablesExecutionError)
-from .chain import Chain
-from .matches.commentmatch import CommentMatch
+from .chain import Chain, BuiltinChain
 from .targets import Target, ChainTarget, UnparsedTarget
 from .deps import get_logger
 
@@ -74,13 +73,15 @@ class IptablesPacketFilterTable:
                                         'POSTROUTING', 'OUTPUT'),
                         'nat' : ('PREROUTING', 'POSTROUTING', 'OUTPUT'),
                         'raw' : ('PREROUTING', 'OUTPUT'),
+                        'security' : ('INPUT', 'OUTPUT', 'FORWARD'),
                     }
 
     __CHAIN_NAME_LIMIT = 28
 
     def __init__(self, table_name: str, chain_prefix: str =None, runner=None):
         """
-        :param table_name: one of ``filter``, ``mangle``, ``nat``, ``raw``
+        :param table_name: one of ``filter``, ``mangle``, ``nat``, ``raw``,
+            or ``security``
         :param chain_prefix: if specified, only chains with names starting
             with this prefix will be included (however, the builtin chains
             are **always** included)
@@ -214,7 +215,7 @@ class IptablesPacketFilterTable:
         """
         return [c for c in self.__chain_map.values() if not c.is_builtin()]
 
-    def get_rules_by_target(self, target: Target):
+    def get_rules_by_target(self, target: Target) -> List[List['ChainRule']]:
         """The return value is a list of lists: the inner list contains
         :class:`ChainRule` objects with the rules of a particular chain that
         are using the specified target, while the outer list corresponds
@@ -278,8 +279,9 @@ class IptablesPacketFilterTable:
         return [chain for rcn, chain in self.__chain_map.items()
                         if rcn.startswith(rcn_prefix)]
 
-    def get_builtin_chain(self, chain_name: str) -> Chain:
-        """Returns the :class:`Chain` object for the specified builtin chain
+    def get_builtin_chain(self, chain_name: str) -> BuiltinChain:
+        """Returns the :class:`BuiltinChain` object with the specified
+        ``chain_name``.
 
         :param chain_name: name of builtin chain
         """
@@ -293,7 +295,7 @@ class IptablesPacketFilterTable:
         """Execute an **iptables(8)** command with the specified arguments.
         The table name is added automatically. For example::
 
-            pft.iptables_run(['-N newchain'])
+            pft.iptables_run(['-N', 'newchain'])
 
         will execute the command::
 
@@ -380,18 +382,6 @@ class IptablesPacketFilterTable:
         """
         self.__unresolved_rules.append(rule)
 
-    @staticmethod
-    def __rule_matches_all_packets(rule: 'ChainRule') -> bool:
-        """Returns True if the rule matches all packets
-        """
-        match_list = rule.get_match_list()
-        if not match_list:
-            return True
-        for match in match_list:
-            if not isinstance(match, CommentMatch):
-                return False
-        return True
-
     def __aggregate_rule_stats(self, chain_name: str, log_level,
                 rule_iter: Iterator['ChainRule'],
                 log_stat_failures: bool) -> Tuple[int, int]:
@@ -415,7 +405,7 @@ class IptablesPacketFilterTable:
             # To alleviate this, we compute the aggregate stats of the
             # remaining rules, and pick the largest value.
             #
-            if self.__rule_matches_all_packets(rule):
+            if rule.matches_all_packets():
                 packet_count_rest, byte_count_rest = \
                         self.__aggregate_rule_stats(chain_name,
                                     logging.INFO, rule_iter, log_stat_failures)
@@ -469,8 +459,14 @@ class IptablesPacketFilterTable:
         return (chain_packet_count, chain_byte_count)
 
     def __calculate_builtin_chain_stats(self, log_stat_failures: bool):
-        """iptables does not report the builtin chain stats, so
-        we calculate them from their rule stats
+        """For all user-chains, we calculate their packet/byte counts
+        based on the packet/byte counts of the rules that direct traffic
+        to those chains.
+        For the builtin chains we need to take a different approach since
+        iptables does not report the builtin chain stats. So we calculate
+        the packet/byte counts from the packet/byte counts of their rules
+        with some heuristics (i.e the packet/byte counts are not guaranteed
+        to be accurate).
         """
         for builtin_chain in self.get_builtin_chains():
             chain_packet_count, chain_byte_count = \
@@ -486,9 +482,9 @@ class IptablesPacketFilterTable:
         """Disassociate from all known chains
         """
         for chain in self.__chain_map.values():
-            chain.clear_pft()
+            chain._clear_pft()          # pylint: disable=protected-access
         for chain in self.__other_chain_map.values():
-            chain.clear_pft()
+            chain._clear_pft()          # pylint: disable=protected-access
 
     def init_from_output(self, iptables_output: str,
                         add_builtins=True, log_parsing_failures=True) -> bool:
@@ -499,7 +495,7 @@ class IptablesPacketFilterTable:
             included in ``subprocess.CompletedProcess.stdout``
         :param add_builtins: if this parameter is ``True`` then for any
             builtin chain that is not in the ``iptables_output`` (or
-            if there is a error parsing its contents), an (empty) chain
+            if there is an error parsing its contents), an (empty) chain
             will be added to the :class:`IptablesPacketFilterTable`
         :param log_parsing_failures: if this parameter is ``True`` then
             parsing failures will be logged; such failures are typically
@@ -539,7 +535,7 @@ class IptablesPacketFilterTable:
                 _logger.warning("Builtin chain not in iptables output: %s",
                                         builtin_name)
             if add_builtins:
-                builtin_chain = Chain(builtin_name, policy='ACCEPT')
+                builtin_chain = BuiltinChain(builtin_name, 'ACCEPT', 0, 0)
                 self.__add_chain(builtin_chain)
         # Update the packet/byte counts
         # This is done after we have resolved all the rules, so that
@@ -617,7 +613,7 @@ class IptablesPacketFilterTable:
             self.__chain_map[rcn] = chain
         else:
             self.__other_chain_map[rcn] = chain
-        chain.set_pft(self)
+        chain._set_pft(self)            # pylint: disable=protected-access
 
     def delete_chain(self, chain: Chain, delete_referring_rules=True) -> None:
         """Delete the specified chain. The chain is flushed first.
@@ -642,7 +638,7 @@ class IptablesPacketFilterTable:
         chain.flush()
         _ = self.iptables_run(['-X', real_chain_name], check=True)
         self.__chain_map.pop(real_chain_name, None)
-        chain.clear_pft()
+        chain._clear_pft()              # pylint: disable=protected-access
 
     def zero_counters(self, chain: Optional[Chain] =None) -> None:
         """Zero the packet and byte counters for the specified chain
