@@ -53,6 +53,8 @@ from linuxnet.iptables import (
                 ConnmarkMatch,
                 LimitMatch,
                 OwnerMatch,
+                PacketMatch,
+                MatchNone,
                 )
 
 root_logger = logging.getLogger()
@@ -951,6 +953,64 @@ Chain prod_OUTPUT (1 references)
         expected_cmd = ['iptables', '-t', 'filter', '-Z']
         self.assertEqual(commands[0].cmd, expected_cmd)
 
+    def test_counters(self):
+        """Test the packet/byte counters
+        """
+        output = """\
+Chain INPUT (policy DROP 10 packets, 500 bytes)
+    pkts      bytes target     prot opt in     out     source               destination
+    500       50000 fw_INPUT   all  --  *      *       0.0.0.0/0            0.0.0.0/0
+
+Chain fw_INPUT (1 references)
+    pkts      bytes target     prot opt in     out     source               destination
+    400       40000 ACCEPT     all  --  *      *       0.0.0.0/0            0.0.0.0/0           state RELATED,ESTABLISHED 
+    100       10000 DROP       all  --  *      *       0.0.0.0/0            0.0.0.0/0
+
+Chain FORWARD (policy DROP 20 packets, 2000 bytes)
+    pkts      bytes target     prot opt in     out     source               destination
+
+Chain OUTPUT (policy ACCEPT 30 packets, 3000 bytes)
+    pkts      bytes target     prot opt in     out     source               destination
+"""
+        pft = IptablesPacketFilterTable('filter')
+        self.assertTrue(pft.init_from_output(output))
+        for bltin in pft.get_builtin_chains():
+            if bltin.get_real_name() == 'INPUT':
+                self.assertEqual(bltin.get_policy(), Targets.DROP)
+                self.assertEqual(bltin.get_policy_packet_count(), 10)
+                self.assertEqual(bltin.get_policy_byte_count(), 500)
+                self.assertEqual(bltin.get_packet_count(), 510)
+                self.assertEqual(bltin.get_byte_count(), 50500)
+            elif bltin.get_real_name() == 'FORWARD':
+                self.assertEqual(bltin.get_policy(), Targets.DROP)
+                self.assertEqual(bltin.get_policy_packet_count(), 20)
+                self.assertEqual(bltin.get_policy_byte_count(), 2000)
+                self.assertEqual(bltin.get_packet_count(), 20)
+                self.assertEqual(bltin.get_byte_count(), 2000)
+            elif bltin.get_real_name() == 'OUTPUT':
+                self.assertEqual(bltin.get_policy(), Targets.ACCEPT)
+                self.assertEqual(bltin.get_policy_packet_count(), 30)
+                self.assertEqual(bltin.get_policy_byte_count(), 3000)
+                self.assertEqual(bltin.get_packet_count(), 30)
+                self.assertEqual(bltin.get_byte_count(), 3000)
+        chain = pft.get_chain_by_rcn('fw_INPUT')
+        self.assertEqual(chain.get_packet_count(), 500)
+        self.assertEqual(chain.get_byte_count(), 50000)
+
+    def test_set_policy(self):
+        """Test setting the policy of a builtin chain
+        """
+        runner = SimulatedIptablesRun(0, None)
+        pft = IptablesPacketFilterTable('filter', runner=runner)
+        init_ok = pft.init_from_output(self.GOOD_OUTPUT)
+        self.assertTrue(init_ok)
+        chain = pft.get_builtin_chain('OUTPUT')
+        chain.set_policy(Targets.DROP)
+        expected_cmd = ['iptables', '-t', 'filter', '-P', 'OUTPUT',
+                                        Targets.DROP.get_target_name()]
+        commands = runner.get_run()
+        self.assertEqual(commands[0].cmd, expected_cmd)
+
 
 class TestChainOperations(unittest.TestCase):
     """Test chain operations. The runner is a no-op
@@ -968,38 +1028,199 @@ class TestChainOperations(unittest.TestCase):
     def test_chain_creation_deletion(self):
         """Create, then delete a chain.
         """
-        try:
-            pft = IptablesPacketFilterTable('filter', runner=self._runner)
-            chain = pft.create_chain('test_chain')
-            pft.delete_chain(chain)
-            result = True
-        except IptablesError:
-            result = False
-        self.assertTrue(result)
+        runner = SimulatedIptablesRun(0, None)
+        pft = IptablesPacketFilterTable('filter', runner=runner)
+        #
+        # Chain creation
+        #
+        chain = pft.create_chain('test_chain')
+        commands = runner.get_run()
+        self.assertEqual(len(commands), 1)
+        self.assertEqual(commands[0].cmd,
+                        ['iptables', '-t', 'filter', '-N', 'test_chain'])
+        runner.clear_run()
+        #
+        # Chain deletion
+        #
+        pft.delete_chain(chain)
+        commands = runner.get_run()
+        self.assertEqual(len(commands), 2)
+        self.assertEqual(commands[0].cmd,
+                        ['iptables', '-t', 'filter', '-F', 'test_chain'])
+        self.assertEqual(commands[1].cmd,
+                        ['iptables', '-t', 'filter', '-X', 'test_chain'])
 
     def test_multi_chain_creation_deletion(self):
         """Create 3 chains, with 2 of them jump'ing to the 3rd.
         Then delete the chain that serves as the terget.
         Rules referencing that chain should be automatically removed.
         """
-        try:
-            pft = IptablesPacketFilterTable('filter', runner=self._runner)
-            chain1 = pft.create_chain('test_chain_1')
-            chain2 = pft.create_chain('test_chain_2')
-            chain3 = pft.create_chain('test_chain_3')
-            chain2.append_rule(ChainRule(target=ChainTarget(chain=chain1)))
-            chain3.append_rule(ChainRule(target=ChainTarget(chain=chain1)))
-            # The following should trigger the automatic deletion of
-            # the rules referencing chain1
-            pft.delete_chain(chain1)
-            self.assertEqual(len(chain2.get_rules()), 0)
-            self.assertEqual(len(chain3.get_rules()), 0)
-            pft.delete_chain(chain2)
-            pft.delete_chain(chain3)
-            result = True
-        except IptablesError:
-            result = False
-        self.assertTrue(result)
+        runner = SimulatedIptablesRun(0, None)
+        pft = IptablesPacketFilterTable('filter', runner=runner)
+        #
+        # Chain creation
+        #
+        chain1 = pft.create_chain('test_chain_1')
+        chain2 = pft.create_chain('test_chain_2')
+        chain3 = pft.create_chain('test_chain_3')
+        commands = runner.get_run()
+        self.assertEqual(len(commands), 3)
+        self.assertEqual(commands[0].cmd,
+                        ['iptables', '-t', 'filter', '-N', 'test_chain_1'])
+        self.assertEqual(commands[1].cmd,
+                        ['iptables', '-t', 'filter', '-N', 'test_chain_2'])
+        self.assertEqual(commands[2].cmd,
+                        ['iptables', '-t', 'filter', '-N', 'test_chain_3'])
+        runner.clear_run()
+        #
+        # Rule creation
+        #
+        chain2.append_rule(ChainRule(target=ChainTarget(chain=chain1)))
+        chain3.append_rule(ChainRule(target=ChainTarget(chain=chain1)))
+        commands = runner.get_run()
+        self.assertEqual(len(commands), 2)
+        self.assertEqual(commands[0].cmd,
+            ['iptables', '-t', 'filter',
+                    '-A', 'test_chain_2', '-j', 'test_chain_1'])
+        self.assertEqual(commands[1].cmd,
+            ['iptables', '-t', 'filter',
+                    '-A', 'test_chain_3', '-j', 'test_chain_1'])
+        runner.clear_run()
+        #
+        # The following should trigger the automatic deletion of
+        # the rules referencing test_chain_1
+        #
+        pft.delete_chain(chain1)
+        commands = runner.get_run()
+        self.assertEqual(len(commands), 4)
+        # Since the order of deletion is not guaranteed, allow both
+        if commands[0].cmd[4] == 'test_chain_2':
+            first_chain = 'test_chain_2'
+            second_chain = 'test_chain_3'
+        else:
+            first_chain = 'test_chain_3'
+            second_chain = 'test_chain_2'
+        self.assertEqual(commands[0].cmd,
+            ['iptables', '-t', 'filter', '-D', first_chain, '1'])
+        self.assertEqual(commands[1].cmd,
+            ['iptables', '-t', 'filter', '-D', second_chain, '1'])
+        self.assertEqual(commands[2].cmd,
+                        ['iptables', '-t', 'filter', '-F', 'test_chain_1'])
+        self.assertEqual(commands[3].cmd,
+                        ['iptables', '-t', 'filter', '-X', 'test_chain_1'])
+        self.assertEqual(len(chain2.get_rules()), 0)
+        self.assertEqual(len(chain3.get_rules()), 0)
+
+    def test_rule_operations(self):
+        """Rule search, deletion
+        """
+        runner = SimulatedIptablesRun(0, None)
+        pft = IptablesPacketFilterTable('filter', runner=runner)
+        #
+        # We create the following structure:
+        #  test_chain_1:
+        #  test_chain_2:
+        #       -p tcp -j test_chain_3
+        #       -j test_chain_1
+        #  test_chain_3:
+        #       -p udp -j test_chain_2
+        #       -j test_chain_1
+        #
+        chain1 = pft.create_chain('test_chain_1')
+        chain2 = pft.create_chain('test_chain_2')
+        chain3 = pft.create_chain('test_chain_3')
+        chain2.append_rule(ChainRule(
+            match=PacketMatch().protocol().equals('tcp')).jump_to(chain=chain3))
+        chain2.append_rule(ChainRule(target=ChainTarget(chain=chain1)))
+        chain3.append_rule(ChainRule(
+            match=PacketMatch().protocol().equals('udp')).jump_to(chain=chain2))
+        chain3.append_rule(ChainRule(target=ChainTarget(chain=chain1)))
+        commands = runner.get_run()
+        self.assertEqual(commands[0].cmd,
+                        ['iptables', '-t', 'filter', '-N', 'test_chain_1'])
+        self.assertEqual(commands[1].cmd,
+                        ['iptables', '-t', 'filter', '-N', 'test_chain_2'])
+        self.assertEqual(commands[2].cmd,
+                        ['iptables', '-t', 'filter', '-N', 'test_chain_3'])
+        self.assertEqual(commands[3].cmd,
+            ['iptables', '-t', 'filter',
+                    '-A', 'test_chain_2', '-p', 'tcp', '-j', 'test_chain_3'])
+        self.assertEqual(commands[4].cmd,
+            ['iptables', '-t', 'filter',
+                    '-A', 'test_chain_2', '-j', 'test_chain_1'])
+        self.assertEqual(commands[5].cmd,
+            ['iptables', '-t', 'filter',
+                    '-A', 'test_chain_3', '-p', 'udp', '-j', 'test_chain_2'])
+        self.assertEqual(commands[6].cmd,
+            ['iptables', '-t', 'filter',
+                    '-A', 'test_chain_3', '-j', 'test_chain_1'])
+        #
+        # Since there are 2 chains that have one rule each jumping
+        # to test_chain_1, the result should be a list containing 2 lists,
+        # with each of those lists containing a single rule.
+        #
+        result = pft.get_rules_by_target(
+                        ChainTarget(real_chain_name='test_chain_1'))
+        self.assertEqual(len(result), 2)
+        for rule_list in result:
+            self.assertEqual(len(rule_list), 1)
+            rule = rule_list[0]
+            owner = rule.get_chain()
+            self.assertTrue(owner is chain2 or owner is chain3)
+        #
+        # Search by match
+        #
+        rules = chain2.find_rule_by(match=MatchNone())
+        self.assertTrue(len(rules), 1)
+        rule = rules[0]
+        self.assertTrue(rule.get_rulenum(), 2)
+        target = rule.get_target()
+        self.assertTrue(isinstance(target, ChainTarget))
+        self.assertTrue(target.get_chain() is chain1)
+        #
+        # Search by match (2)
+        #
+        rules = chain3.find_rule_by(
+                        match=PacketMatch().protocol().equals('udp'))
+        self.assertTrue(len(rules), 1)
+        rule = rules[0]
+        self.assertTrue(rule.get_rulenum(), 1)
+        target = rule.get_target()
+        self.assertTrue(isinstance(target, ChainTarget))
+        self.assertTrue(target.get_chain() is chain2)
+        #
+        # Search by target
+        #
+        rules = chain3.find_rule_by(target=ChainTarget(chain=chain1))
+        self.assertTrue(len(rules), 1)
+        rule = rules[0]
+        self.assertTrue(rule.get_rulenum(), 2)
+        target = rule.get_target()
+        self.assertTrue(isinstance(target, ChainTarget))
+        self.assertTrue(target.get_chain() is chain1)
+        #
+        # Search by match and target
+        #
+        rules = chain3.find_rule_by(
+                        match=PacketMatch().protocol().equals('udp'),
+                        target=ChainTarget(chain=chain2))
+        self.assertTrue(len(rules), 1)
+        rule = rules[0]
+        self.assertTrue(rule.get_rulenum(), 1)
+        target = rule.get_target()
+        self.assertTrue(isinstance(target, ChainTarget))
+        self.assertTrue(target.get_chain() is chain2)
+        runner.clear_run()
+        #
+        # Delete the first rule of test_chain_3
+        #
+        n_deleted = chain3.delete_rule_if(
+                        match=PacketMatch().protocol().equals('udp'),
+                        target=ChainTarget(chain=chain2))
+        self.assertTrue(n_deleted, 1)
+        commands = runner.get_run()
+        self.assertTrue(commands[0].cmd,
+                ['iptables', '-t', 'filter', '-D', 'test_chain_3', 1])
 
     def test_chain_rule_ownership(self):
         """Attempt to insert a rule into a chain twice.
@@ -1009,24 +1230,14 @@ class TestChainOperations(unittest.TestCase):
         chain = pft.create_chain('test_chain')
         rule = ChainRule(target=Targets.ACCEPT)
         chain.append_rule(rule)
-        try:
-            chain.append_rule(rule)
-            inserted = True
-        except IptablesError:
-            inserted = False
-        self.assertFalse(inserted)
+        self.assertRaises(IptablesError, lambda: chain.append_rule(rule))
         # Remove the rule, verify it now has no owner
         chain.flush()
-        self.assertEqual(rule.get_chain(), None)
-        try:
-            chain.append_rule(rule)
-            inserted = True
-        except IptablesError:
-            inserted = False
-        self.assertTrue(inserted)
+        self.assertTrue(rule.get_chain() is None)
+        chain.append_rule(rule)
         # Delete it by rule number
         chain.delete_rulenum(rule.get_rulenum())
-        self.assertEqual(rule.get_chain(), None)
+        self.assertTrue(rule.get_chain() is None)
 
     def test_chain_rule_numbering(self):
         """Verify rule numbering when rules are inserted/deleted.
